# Copyright (c) Alibaba, Inc. and its affiliates.

from odps.udf import BaseUDAF
from odps.udf import annotate


@annotate('*->array<string>')
class CountCatesKVS(BaseUDAF):
    """Count category feature as key-value map."""

    def new_buffer(self):
        """Local buffer.

        buffer[0]: the number of maximum keys to be kept.
        buffer[1]: a list of event types to be counted.
            e.g., ['event1','event2',...,'eventn|eventm']]
        buffer[2]: event processing formulas.
            e.g., ['event2/event1','eventn|eventm/event2',...]
        buffer[3]: event statistics dict.
            e.g., {'event1':[{cate1 stats dict},{cate2 stats dict},..],
                   'event2':[{cate1 stats dict},{cate2 stats dict}...],...}
        """
        return [-1, list(), list(), dict(), False]

    def iterate(self,
                buffer,
                cates,
                values,
                event,
                event_types,
                event_formulas='',
                sep='\x1d',
                top_k=50,
                is_punish=False):
        """Iterate category line and put into buffer.

        Args:
            buffer: udaf buffer.
            cates: a category or a list of categories to be counted.
            values: a number or a list of numbers to be accumulated.
            event: event field.
            event_types: a event_type or a list of event type, combined events are separated with |.
            event_formulas: a list of calculating formula for two events.
            sep: separator for multi-value categories.
            top_k: the number of maximum keys to be kept.
            is_punish: whether to pubish denominator with mean value when calculating the formula.
        """
        assert type(values) in (int, float, list), 'type of values must be int or double or list.'
        if isinstance(values, list):
            for ele in values:
                assert type(ele) in (int, float), 'values element type must be bigint or double'
        assert type(event_types) in (str, list), 'type of event_types must be str or list.'
        assert type(event_formulas) in (str, list), 'type of event_formulas must be str or list.'
        assert top_k is None or top_k > 0, 'top_k must greater than 0'

        if not isinstance(cates, list):
            cates = [cates]
        if not isinstance(values, list):
            values = [values for _ in range(len(cates))]
        if not isinstance(event_types, list):
            event_types = [event_types]
        if not isinstance(event_formulas, list):
            event_formulas = [event_formulas] if len(event_formulas) > 0 and event_formulas is not None else []

        for event_formula in event_formulas:
            form_events = event_formula.split('/')
            assert len(form_events) == 2, 'now only div formula is supported'
            assert form_events[0] in event_types and form_events[1] in event_types

        buffer[0] = top_k
        buffer[1] = event_types
        buffer[2] = event_formulas
        buffer[4] = is_punish
        event_stats_dict = buffer[3]

        is_in_event, event_type = self.in_event(event, event_types)
        if is_in_event:
            cates_kvs = event_stats_dict.setdefault(event_type, list())
            if len(cates_kvs) == 0:
                cates_kvs = [dict() for _ in range(len(cates))]
            for i, (cate, v, tmp_dict) in enumerate(zip(cates, values, cates_kvs)):
                if cate is not None and (len(str(cate).strip()) > 0):
                    cate = str(cate)
                    if sep is not None:
                        tags = [tag.strip() for tag in cate.split(sep)]
                    else:
                        tags = [cate]
                    for tag in tags:
                        if tag in tmp_dict:
                            tmp_dict[tag] = tmp_dict[tag] + v
                        else:
                            tmp_dict[tag] = v
                    cates_kvs[i] = tmp_dict
                else:
                    cates_kvs[i] = tmp_dict
            event_stats_dict[event_type] = cates_kvs
        buffer[3] = event_stats_dict

    def merge(self, buffer, pbuffer):
        """Merge multiple machine's pbuffer into buffer."""
        buffer[0] = pbuffer[0]
        buffer[1] = pbuffer[1]
        buffer[2] = pbuffer[2]
        buffer[4] = pbuffer[4]
        merged_event_stats_dict = buffer[3]
        for event, pcates_kvs in pbuffer[3].items():
            cates_kvs = buffer[3].setdefault(event, list())
            if len(cates_kvs) == 0:
                cates_kvs = [dict() for _ in range(len(pcates_kvs))]
            merged_cates_kvs = []
            for pbuffer_dict, buffer_dict in zip(pcates_kvs, cates_kvs):
                for k, v in pbuffer_dict.items():
                    if k in buffer_dict:
                        buffer_dict[k] = buffer_dict[k] + v
                    else:
                        buffer_dict[k] = v
                merged_cates_kvs.append(buffer_dict)
            merged_event_stats_dict[event] = merged_cates_kvs
        buffer[3] = merged_event_stats_dict

    def terminate(self, buffer):
        """Convert buffer to key-value format."""
        top_k = buffer[0]
        event_types = buffer[1]
        event_formulas = buffer[2]
        event_stats_dict = buffer[3]
        is_punish = buffer[4]

        cates_len = 0
        for event, cates_kvs in event_stats_dict.items():
            if len(cates_kvs) > 0:
                cates_len = len(cates_kvs)
                break

        # Apply formula
        if len(event_formulas) > 0:
            for event_formula in event_formulas:
                # now only div formula is supported
                formula_events = event_formula.split('/')
                event_l, event_r = formula_events[0], formula_events[1]

                if event_l in event_stats_dict and event_r in event_stats_dict:
                    cate_kvs_l = event_stats_dict[event_l]
                    cate_kvs_r = event_stats_dict[event_r]
                    event_stats_dict[event_formula] = self.divide_cate_kvs(cate_kvs_l, cate_kvs_r, is_punish)
                else:
                    event_stats_dict[event_formula] = [dict() for _ in range(cates_len)]

        # Caculate Topk
        for key, cates_kvs in event_stats_dict.items():
            for i, tmp_dict in enumerate(cates_kvs):
                items = sorted(tmp_dict.items(), key=lambda x: x[1], reverse=True)
                if len(items) > top_k:
                    items = items[:top_k]
                tmp_dict = {x: y for x, y in items}
                cates_kvs[i] = tmp_dict
            event_stats_dict[key] = cates_kvs

        # Format as result
        result = []
        for key in event_types + event_formulas:
            if key in event_stats_dict:
                cates_kvs = event_stats_dict[key]
                for tmp_dict in cates_kvs:
                    tmp = []
                    for k, v in sorted(tmp_dict.items()):
                        tmp.append('{_k}:{_v}'.format(_k=k, _v=v))
                    if len(tmp) > 0:
                        result.append(chr(29).join(tmp))
                    else:
                        result.append('')
            else:
                result.extend([''] * cates_len)

        return result

    def in_event(self, event, event_types):
        """Checks event in event_types or not."""
        for event_type in event_types:
            if ('|' in event_type and event in event_type.split('|')) or (event_type == event):
                return True, event_type
        return False, ''

    def divide_cate_kvs(self, cates_kvs_left, cates_kvs_right, is_punish=False):
        """Divide two event category key-values.

        Args:
            cates_kvs_left: category key-values of left event
            cates_kvs_right: category key-values of right event
            is_punish: whether to pubish denominator with mean value when calculating the formula.
        """
        cates_kvs = []
        for kv_l, kv_r in zip(cates_kvs_left, cates_kvs_right):
            tmp_dict = dict()
            punish = 0.0
            if is_punish:
                punish = sum(kv_r.values()) / len(kv_r)
            if len(kv_l) > 0 and len(kv_r) > 0:
                for k, v_l in kv_l.items():
                    if k in kv_r:
                        v_r = kv_r[k]
                        tmp_dict[k] = round(v_l / (v_r * 1.0 + punish), 4)
            cates_kvs.append(tmp_dict)
        return cates_kvs
